#include "cssysdef.h"
#include <iostream.h>

#include "xCoDS.h"


/**
 * The implementation of xCoDS
 */

xCoDS::xCoDS(int nCol) {
    nColliders = nCol;
    ColliderArray = new (xCollider*)[nCol];
    OxProj = new (xCoord*)[2*nCol];
    OyProj = new (xCoord*)[2*nCol];
    OzProj = new (xCoord*)[2*nCol];
    CollisionCache = new (bool*)[nCol];
    for(int i=0;i<nCol;i++)
      CollisionCache[i] = new bool[nCol];
}

xCoDS::~xCoDS() {
  int i;
  for(i=0;i<nColliders;i++) {
    delete ColliderArray[i];
    delete OxProj[2*i];
    delete OxProj[2*i+1];
    delete OyProj[2*i];
    delete OyProj[2*i+1];
    delete OzProj[2*i];
    delete OzProj[2*i+1];
  }
  delete [] ColliderArray;
  delete [] OxProj;
  delete [] OyProj;
  delete [] OyProj;
  for(i=0;i<nColliders;i++)
    delete [] CollisionCache[i];
  delete [] CollisionCache;

}

void xCoDS::setCollider(int n, xCollider* pCol) {
    ColliderArray[n] = pCol;
    // set the coordinates for this collider
    pCol->xAABB[0] = OxProj[2*n] =  new xCoord(pCol->OxLeft(),true,n);
    pCol->xAABB[1] = OxProj[2*n+1] =  new xCoord(pCol->OxRight(),false,n);
    pCol->xAABB[2] = OyProj[2*n] =  new xCoord(pCol->OyLeft(),true,n);
    pCol->xAABB[3] = OyProj[2*n+1] =  new xCoord(pCol->OyRight(),false,n);
    pCol->xAABB[4] = OzProj[2*n] =  new xCoord(pCol->OzLeft(),true,n);
    pCol->xAABB[5] = OzProj[2*n+1] =  new xCoord(pCol->OzRight(),false,n);
    // construct the PCO set for the this collider
    pCol->buildPCO(nColliders);
}

void xCoDS::resort(short whichAxis) {
  xCoord** axis;
  switch(whichAxis) {
  case AXIS_OX: axis = OxProj; break; 
  case AXIS_OY: axis = OyProj; break; 
  case AXIS_OZ: axis = OzProj; break; 
  }
  
    // insertion sort
    int i;
    for(i=1;i<2*nColliders;i++) {
	if(axis[i]->Value<axis[i-1]->Value) {
	    // here we need to reinsert element i between 0 and i
	    xCoord* aux = axis[i];
	    int aux_i = axis[i]->Collider_i;
	    int j=i-1;
	    do {
	      int crt_i = axis[j]->Collider_i;
	      // here we should track the l-r overlaps
	      if(aux->isLeft()) {
		// HAVE TO USE INTS and code axes 1,2,4
		// the box stretches over the other -> PCO
		
		ColliderArray[aux_i]->PCO[crt_i] |= whichAxis;
		ColliderArray[crt_i]->PCO[aux_i] |= whichAxis;		
		
	      } else {
		// aux is a right margin
		if(axis[j]->isLeft()) { // not overlapping anymore
		
		  ColliderArray[aux_i]->PCO[crt_i] &= ~whichAxis;
		  ColliderArray[crt_i]->PCO[aux_i] &= ~whichAxis;
		
		}
	      }
	      
	      axis[j+1]=axis[j];
	      j--;
	    } while((j>=0) && (aux->Value<axis[j]->Value));
	    // finally place the element at the right position
	    axis[j+1]=aux;
	}
    }
}

void xCoDS::spit(xCoord* axis[]) {
  int i;
  for(i=0;i<2*nColliders;i++) {
    cout<<" "<<axis[i]->Value<<"("<<axis[i]->Collider_i<<")";
    ColliderArray[axis[i]->Collider_i]->printPCO(nColliders);
  }
  cout<<"\n";
}



int xCoDS::updateCollisions(iCollideSystem* cds, int& pcollisions) {

  // resort the axes first
  resort(AXIS_OX);
  resort(AXIS_OY);
  resort(AXIS_OZ);

  int i,j;
  int pcodetected = 0;
  for(i=0;i<(nColliders-1);i++)
    // for each body check its PCO
    for(j=i+1;j<nColliders;j++) {
      if(ColliderArray[i]->Moved()||ColliderArray[j]->Moved()) {
	// either moved, so we can't relly on cache
	if(ColliderArray[i]->isPotentiallyColliding(j)) {
	  // cout<<" "<<i<<" - "<<j;
	  pcollisions++;
	  // they might collide!
	  // have to call RAPID for this pair
	  // CALL RAPID AND SET THE CACHE ACCORDINGLY
	  // NEW: let the objects handle the collision
	  //      by calling RAPID or other lower level
	  xObject* pOi = ColliderArray[i]->getObject();
	  xObject* pOj = ColliderArray[j]->getObject();
	  bool cij = pOi->isCollidingWith(cds, pOj);
	  CollisionCache[i][j] = CollisionCache[j][i] = cij;
	  if(cij) pcodetected++;
	} else {
	  // they don't collide so set the collision cache to false
	  CollisionCache[i][j] = CollisionCache[j][i] = false;
	}
      }
	  
    }

  // done updating, now unset all movement flags!
  for(i=0;i<nColliders;i++)
    ColliderArray[i]->setMovedFlag(false);
  return pcodetected;
}
