#ifndef __XCODS_H__
#define __XCODS_H__

#include "xCoord.h"
#include "xCollider.h"

class iCollideSystem;

#define AXIS_OX 1
#define AXIS_OY 2
#define AXIS_OZ 4

/**
 * the n-body collision detection system
 */

class xCoDS {
    /**
     * The number of colliders involved
     */
    int nColliders;

    /**
     * The array with the involved colliders
     */
    xCollider** ColliderArray;


    /**
     * This is the collision cache... it only keeps
     * the collisions detected with RAPID
     */
    bool** CollisionCache;

 public:
    /**
     * The projections of AABBs on the 3 axis
     */
    xCoord** OxProj;
    xCoord** OyProj;
    xCoord** OzProj;

 public:
    /**
     * Constructor
     * @param nCol the number of colliders supervised by the system
     */
    xCoDS(int nCol);

 
    /**
     * Destructor
     */
    ~xCoDS();

    /**
     * Sets the index n in the collider array to point to pCol
     */
    void setCollider(int n, xCollider* pCol);
    xCollider* getCollider(int n) { return ColliderArray[n]; };


    /**
     * Re-sorts one of the given axes
     */
    void resort(short whichAxis);



    /**
     * updates the collision list
     * - takes into account the collision cache
     *   for the bodies that did not move
     * - for pairs of colliders which moved
     *   and are in PCO, call RAPID
     * returns the number of collisions and in the int
     * the number of potential collisions
     */
    int updateCollisions(iCollideSystem*, int&);
    

    /**
     * good for debug. prints the index of the colliders
     * on an axis
     */
    void spit(xCoord* axis[]);
    
};

#endif


